#include "bpop_main.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

GameScene currentScene = GAME_SCENE_MENU;
GameScene nextScene = GAME_SCENE_NONE;

GameDifficulty gameDifficulty = GAME_DIFFICULTY_NORMAL;

Storage storage = {
    .version = 1,
};

int outlineSizeLoc = 0;
int outlineColorLoc = 0;
int textureSizeLoc = 0;

Camera3D camera = { 
    .position = { 0.0f, 0.0f, -10.0f },
    .target = { 0.0f, 0.0f, 0.0f },
    .up = { 0.0f, 1.0f, 0.0f },
    .projection = CAMERA_ORTHOGRAPHIC,
    .fovy = 320.0f,
};


Playfield playfield = {
    .fieldSize = {250, 300},
    .waterLevel = 200.0f,
    .minHeight = 90.0f,
    .maxHeight = 280.0f,
};

GameTime gameTime = {
    .fixedDeltaTime = 1.0f / 60.0f,
};

Color bubbleTypeColors[] = {
    COLOR_FROM_HEX(0xaaccff),
    COLOR_FROM_HEX(0xaa7733),
    COLOR_FROM_HEX(0x33ffaa),
    COLOR_FROM_HEX(0xff33aa),
    COLOR_FROM_HEX(0xaa33ff),
};

Model bubbleModel;
Shader bubbleOutlineShader;
RenderTexture2D bubbleFieldTexture;

int isClickActionBlocked = 0;

float BubbleLevelRadius(int level)
{
    return powf((level + 1) * 120, .5f);
}

int IsClickActioned()
{
    int result = 0;
    if (IsMouseButtonReleased(MOUSE_LEFT_BUTTON) && !isClickActionBlocked)
    {
        result = 1;
    }
    return result;
}

int Button(const char *text, Vector2 position, Vector2 size)
{
    int result = 0;
    Rectangle rect = {position.x, position.y, size.x, size.y};
    if (CheckCollisionPointRec(GetMousePosition(), rect))
    {
        if (IsClickActioned())
        {
            result = 1;
        }
    }
    DrawRectangleRec(rect, (Color){200, 200, 200, 255});
    DrawRectangleLinesEx(rect, 2, BLACK);
    float width = MeasureText(text, 20);
    DrawText(text, position.x + size.x * 0.5f - width * 0.5f, position.y + 10, 20, BLACK);
    return result;
}

void PlayfieldFixedUpdate(Playfield *playfield)
{
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        bubble->sameLevelContact = 0;
    }

    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (!bubble->flagIsActive) continue;
        float r = bubble->radius;

        for (int j = i + 1; j < MAX_BUBBLES; j++)
        {
            Bubble *other = &playfield->bubbles[j];
            if (!other->flagIsActive) continue;
            float otherR = other->radius;
            float sumR2 = (r + otherR) * (r + otherR);
            float d2 = Vector2DistanceSqr(bubble->position, other->position);
            int canMerge = bubble->bubbleLevel == other->bubbleLevel && bubble->bubbleType == other->bubbleType;
            if (d2 < sumR2 * 1.05f)
            {
                if (canMerge)
                {
                    bubble->sameLevelContact = 1;
                    other->sameLevelContact = 1;
                }
                if (canMerge && bubble->bubbleMergeCooldown <= 0.0f 
                    && other->bubbleMergeCooldown <= 0.0f)
                {
                    // merge bubbles
                    bubble->bubbleLevel++;
                    bubble->radius = BubbleLevelRadius(bubble->bubbleLevel);
                    bubble->bubbleMergeCooldown = 1.0f;
                    other->flagIsActive = 0;
                }
            }

            if (d2 < sumR2)
            {
                float overlap = r + otherR - sqrtf(d2);
                Vector2 normal = Vector2Normalize(Vector2Subtract(other->position, bubble->position));
                // resolve overlap by moving the bubbles apart
                const float errorCorrection = 0.25f;
                bubble->position = Vector2Subtract(bubble->position, Vector2Scale(normal, overlap * errorCorrection));
                other->position = Vector2Add(other->position, Vector2Scale(normal, overlap * errorCorrection));

                // bounce off each other
                Vector2 relativeVelocity = Vector2Subtract(bubble->velocity, other->velocity);
                float dot = Vector2DotProduct(relativeVelocity, normal);
                if (dot > 0.0f)
                {
                    // DrawLineV(bubble->position, other->position, COLOR_FROM_HEX(0xff0000));
                    float impulse = -dot * 0.85f;
                    bubble->velocity = Vector2Add(bubble->velocity, Vector2Scale(normal, impulse));
                    other->velocity = Vector2Subtract(other->velocity, Vector2Scale(normal, impulse));
                }
            }
        }

        if (!bubble->sameLevelContact)
        {
            bubble->bubbleMergeCooldown = 1.0f;
        }
        else
        {
            bubble->bubbleMergeCooldown -= gameTime.fixedDeltaTime;
        }

        float buoyancy = -20.0f;
        if (bubble->position.y < playfield->waterLevel)
        {
            buoyancy = (playfield->waterLevel - bubble->position.y) * 0.5f;
        }

        int isOutsideZone = bubble->position.y < playfield->minHeight + bubble->radius || 
            bubble->position.y > playfield->maxHeight - bubble->radius;
        bubble->lifeTime += gameTime.fixedDeltaTime;
        bubble->isOutsideZone = isOutsideZone && bubble->lifeTime > 1.0f;

        bubble->velocity = Vector2Add(bubble->velocity, (Vector2){0, buoyancy});
        bubble->velocity = Vector2Scale(bubble->velocity, 0.92f);
        bubble->position.x += bubble->velocity.x * gameTime.fixedDeltaTime;
        bubble->position.y += bubble->velocity.y * gameTime.fixedDeltaTime;
        if ((bubble->position.x < r && bubble->velocity.x < 0.0f) || 
            (bubble->position.x > playfield->fieldSize.x - r && bubble->velocity.x > 0.0f))
        {
            bubble->velocity.x *= -0.9f;
        }

        bubble->position.x = (bubble->position.x < r) ? r : (bubble->position.x > playfield->fieldSize.x - r) ? playfield->fieldSize.x - r : bubble->position.x;
        // bubble->position.y = (bubble->position.y < r) ? r : (bubble->position.y > playfield->fieldSize.y - r) ? playfield->fieldSize.y - r : bubble->position.y;

        // debug velocity
        // DrawLineV(bubble->position, Vector2Add(bubble->position, Vector2Scale(bubble->velocity, 1.0f)), COLOR_FROM_HEX(0xff0000));
    }

    int outsideCount = 0;
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (bubble->isOutsideZone)
        {
            outsideCount++;
        }
    }

    if (outsideCount == 0)
    {
        playfield->strikes = 0;
        playfield->gameoverCooldown = 0.0f;
    }
    if (playfield->strikes >= MAX_STRIKES)
    {
        playfield->gameoverCooldown += gameTime.fixedDeltaTime;
        if (playfield->gameoverCooldown > GAMEOVER_COOLDOWN)
        {
            nextScene = GAME_SCENE_GAMEOVER;
        }
    }
}

void PlayfieldTryAddBubble(Playfield *playfield, Vector2 position)
{
    if (playfield->strikes >= MAX_STRIKES) return;

    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (!bubble->flagIsActive)
        {
            bubble->flagIsActive = 1;
            bubble->position = position;
            bubble->velocity = (Vector2){GetRandomValue(-100, 100), GetRandomValue(-100, 100)};
            bubble->bubbleType = playfield->nextBubbleType;
            bubble->bubbleLevel = playfield->nextBubbleLevel;
            bubble->radius = BubbleLevelRadius(bubble->bubbleLevel);
            bubble->lifeTime = 0.0f;
            playfield->spawnedBubbleCount++;

            for (int j = 0; j < MAX_BUBBLES; j += 1)
            {
                Bubble *other = &playfield->bubbles[j];
                if (!other->flagIsActive) continue;
                if (other->isOutsideZone)
                {
                    playfield->strikes++;
                    break;
                }
            }

            playfield->nextBubbleType = GetRandomValue(0, gameDifficulty);
            playfield->nextBubbleLevel = GetRandomValue(0, 3);
            break;
        }
    }

}

PlayfieldScores CalculatePlayfieldScores(Playfield *playfield)
{
    PlayfieldScores scores = {0};
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (bubble->flagIsActive)
        {
            scores.bubbleCount++;
            uint32_t bubbleScore = 1 << bubble->bubbleLevel;
            scores.score += bubbleScore;

            if (bubble->isOutsideZone)
            {
                scores.outsideBubbleCount++;
            }
        }
    }
    scores.score += playfield->spawnedBubbleCount;
    return scores;
}

Vector2 PlayfieldPositionToSpawnPosition(Playfield *playfield, Vector2 position)
{
    Vector2 spawnPosition = position;
    spawnPosition.y = BubbleLevelRadius(5);
    return spawnPosition;
}

Vector2 PlayfieldScreenToSpawnPosition(Playfield *playfield, Camera3D camera, Vector2 screenPosition)
{
    Vector3 cursorPosition = GetScreenToWorldRay(screenPosition, camera).position;
    cursorPosition.x += playfield->fieldSize.x / 2;
    cursorPosition.y += playfield->fieldSize.y / 2;

    Vector2 pos = {cursorPosition.x, cursorPosition.y};
    return PlayfieldPositionToSpawnPosition(playfield, pos);
}

void DrawBubble(Vector3 position, int level, Color color)
{
    float bubbleExtraRadius = 5.0f;
    float r = BubbleLevelRadius(level) + bubbleExtraRadius;
    DrawModel(bubbleModel, position, r, color);
    if (level < 1) return;
    position.z -= r;
    float tinyR = level < 6 ? 2 : 4;
    int count = level < 6 ? level : level - 5;
    for (int i = 0; i < count; i++)
    {
        float ang = (i * 25.0f + 30.0f) * DEG2RAD;
        float offsetR = i % 2 == 0 ? 0.4f : 0.7f;
        Vector3 offset = {cosf(ang) * offsetR * r, sinf(ang) * offsetR * r, 0};
        DrawModel(bubbleModel, Vector3Add(position, offset), tinyR, WHITE);
    }
}

void PlayfieldDrawBubbles(Playfield *playfield, Camera3D camera)
{
    DrawCube((Vector3){0, 0, 0}, playfield->fieldSize.x, playfield->fieldSize.y, 0, COLOR_FROM_HEX(0xbbddff));
    DrawCube((Vector3){0, (playfield->waterLevel - playfield->fieldSize.y) * 0.5f, 0}, playfield->fieldSize.x, 
        playfield->waterLevel, 0, COLOR_FROM_HEX(0x225588));
    
    // cursor bubble
    if (currentScene == GAME_SCENE_PLAY)
    {
        Vector2 mousePos = GetMousePosition();
        Vector2 spawnPosition = PlayfieldScreenToSpawnPosition(playfield, camera, mousePos);
        Vector3 drawPos = (Vector3){spawnPosition.x - playfield->fieldSize.x * 0.5f, spawnPosition.y - playfield->fieldSize.y * 0.5f, 0};
        if (playfield->strikes < MAX_STRIKES && drawPos.x >= -playfield->fieldSize.x * 0.5f && drawPos.x <= playfield->fieldSize.x * 0.5f)
        {
            DrawBubble(drawPos, playfield->nextBubbleLevel, bubbleTypeColors[playfield->nextBubbleType]);
        }
    }

    // DrawRectangle(0, 0, playfield->fieldSize.x, playfield->fieldSize.y, COLOR_FROM_HEX(0x225588));
    rlPushMatrix();
    rlTranslatef(-playfield->fieldSize.x / 2, -playfield->fieldSize.y / 2, 0);
    // draw bubbles into playfield space
    float blink = sinf(gameTime.time * 10.0f) * 0.2f + 0.5f;
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (!bubble->flagIsActive) continue;
        Vector3 position = {bubble->position.x, bubble->position.y, 0};
        Color bubbleColor = bubbleTypeColors[bubble->bubbleType];
        int isOutsideZone = bubble->isOutsideZone;
        
        if (isOutsideZone)
        {
            bubbleColor = ColorLerp(bubbleColor, COLOR_FROM_HEX(0xff4433), blink);
        }
        // lazy: encode id into rgb values
        bubbleColor.r = bubbleColor.r - i * 8 % 32;
        bubbleColor.g = bubbleColor.g - (i * 8 / 32) % 32;
        bubbleColor.b = bubbleColor.b - (i * 8 / 512) % 32;
        bubbleColor.a = 255;
        DrawBubble(position, bubble->bubbleLevel, bubbleColor);
    }
    rlPopMatrix();
}

void PlayfieldDrawRange(Playfield *playfield, Camera3D camera)
{
    Color rangeLimitColor = COLOR_FROM_HEX(0xff4400);
    int divides = 10;
    float divWidth = playfield->fieldSize.x / divides;
    for (int i = 0; i < divides; i+=2)
    {
        float x = i * divWidth - playfield->fieldSize.x * 0.5f + divWidth * 1.0f;
        DrawCube((Vector3){x, playfield->minHeight - playfield->fieldSize.y * 0.5f, 0}, divWidth, 5, 5, rangeLimitColor);
        DrawCube((Vector3){x, playfield->maxHeight - playfield->fieldSize.y * 0.5f, 0}, divWidth, 5, 5, rangeLimitColor);
    }
}

void PlayfieldFullDraw(Playfield *playfield, Camera3D camera)
{
    ClearBackground(COLOR_FROM_HEX(0x4488cc));

    BeginTextureMode(bubbleFieldTexture);
    rlSetClipPlanes(-128.0f, 128.0f);
    BeginMode3D(camera);

    ClearBackground(BLANK);
    PlayfieldDrawBubbles(playfield, camera);
    EndMode3D();
    EndTextureMode();

    float outlineSize = 1.0f;
    float outlineColor[4] = { 1.0f, 1.0f, 1.0f, 1.0f };
    float textureSize[2] = { (float)bubbleFieldTexture.texture.width, (float)bubbleFieldTexture.texture.height };

    SetShaderValue(bubbleOutlineShader, outlineSizeLoc, &outlineSize, SHADER_UNIFORM_FLOAT);
    SetShaderValue(bubbleOutlineShader, outlineColorLoc, outlineColor, SHADER_UNIFORM_VEC4);
    SetShaderValue(bubbleOutlineShader, textureSizeLoc, textureSize, SHADER_UNIFORM_VEC2);

    rlDisableDepthMask();
    BeginShaderMode(bubbleOutlineShader);
    DrawTexturePro(bubbleFieldTexture.texture, (Rectangle){0, 0, (float)bubbleFieldTexture.texture.width, 
        -(float)bubbleFieldTexture.texture.height}, (Rectangle){0, 0, (float)GetScreenWidth(), (float)GetScreenHeight()}, 
        (Vector2){0, 0}, 0.0f, WHITE);
    EndShaderMode();
    rlEnableDepthMask();

    BeginMode3D(camera);
    PlayfieldDrawRange(playfield, camera);
    EndMode3D();

    const char *difficultyText = "Tutorial";
    switch (gameDifficulty)
    {
    case GAME_DIFFICULTY_EASY: difficultyText = "Easy"; break;
    case GAME_DIFFICULTY_NORMAL: difficultyText = "Normal"; break;
    case GAME_DIFFICULTY_HARD: difficultyText = "Hard"; break;
    default:
        break;
    }
    const char *modeText = TextFormat("Mode: %s", difficultyText);
    int screenWidth = GetScreenWidth();
    int x = screenWidth - 215;
    DrawText(modeText, x, 10, 20, WHITE);
    DifficultyScores table = storage.scores[gameDifficulty];
    for (int i = 0; i < 8; i++)
    {
        HighscoreEntry entry = table.highscores[i];
        if (entry.score == 0) break;
        char buffer[64];
        sprintf(buffer, "%d:", i + 1);
        int y = 40 + i * 30;
        DrawText(buffer, x + 18 - MeasureText(buffer, 20), y, 20, WHITE);
        sprintf(buffer, "%d", entry.score);
        DrawText(buffer, x + 55 - MeasureText(buffer, 20) / 2, y, 20, WHITE);
        sprintf(buffer, "%s", entry.date);
        DrawText(buffer, screenWidth - 15 - MeasureText(buffer, 20), y, 20, WHITE);
    }
}

void UpdateSceneGameOver()
{
    if (IsKeyPressed(KEY_ENTER))
    {
        nextScene = GAME_SCENE_MENU;
    }
    // Draw
    //----------------------------------------------------------------------------------

    ClearBackground(COLOR_FROM_HEX(0x4488cc));

    PlayfieldFullDraw(&playfield, camera);

    DrawText("Game Over", 20, 20, 35, WHITE);
    
    PlayfieldScores scores = CalculatePlayfieldScores(&playfield);
    DrawText(TextFormat("Final Score: %d", scores.score), 20, 90, 20, WHITE);

    if (Button("Restart", (Vector2){20, 130}, (Vector2){100, 50}))
    {
        nextScene = GAME_SCENE_MENU;
    }
}

void UpdateScenePlay()
{
    if (IsKeyPressed(KEY_ESCAPE))
    {
        nextScene = GAME_SCENE_MENU;
    }

    if (IsClickActioned() && playfield.strikes < MAX_STRIKES)
    {
        Vector2 pos = PlayfieldScreenToSpawnPosition(&playfield, camera, 
            GetMousePosition());
        if (pos.y >= 0.0f && pos.y <= playfield.fieldSize.y
            && pos.x >= 0.0f && pos.x <= playfield.fieldSize.x)
        {
            PlayfieldTryAddBubble(&playfield, pos);
        }
    }
    
    while (gameTime.fixedTime < gameTime.time)
    {
        gameTime.fixedTime += gameTime.fixedDeltaTime;
        PlayfieldFixedUpdate(&playfield);
    }

    // Draw
    //----------------------------------------------------------------------------------

    PlayfieldFullDraw(&playfield, camera);

    PlayfieldScores scores = CalculatePlayfieldScores(&playfield);
    DrawText(TextFormat("Score: %d", scores.score), 10, 10, 20, WHITE);
    DrawText(TextFormat("Bubbles: %d", scores.bubbleCount), 10, 40, 20, WHITE);
    DrawText(TextFormat("Spawned: %d", playfield.spawnedBubbleCount), 10, 70, 20, WHITE);
    DrawText(TextFormat("Outside: %d", scores.outsideBubbleCount), 10, 100, 20, WHITE);
    DrawText(TextFormat("Strikes: %d", playfield.strikes), 10, 130, 20, WHITE);
    DrawText(TextFormat("Gameover in: %.1f", GAMEOVER_COOLDOWN - playfield.gameoverCooldown), 10, 160, 20, WHITE);
}

void UpdateSceneMenu()
{
    // Draw
    //----------------------------------------------------------------------------------

    ClearBackground(COLOR_FROM_HEX(0x4488cc));

    DrawText("Bubble Pop", GetScreenWidth() / 2 - MeasureText("Bubble Pop", 40) / 2, 10, 40, WHITE);
    
    int y = 100;
    if (Button("Tutorial", (Vector2){GetScreenWidth() / 2 - 100, y}, (Vector2){200, 50}))
    {
        gameDifficulty = GAME_DIFFICULTY_TUTORIAL;
        nextScene = GAME_SCENE_PLAY;
    }

    if (Button("Play easy", (Vector2){GetScreenWidth() / 2 - 100, y += 60}, (Vector2){200, 50}))
    {
        gameDifficulty = GAME_DIFFICULTY_EASY;
        nextScene = GAME_SCENE_PLAY;
    }

    if (Button("Play normal", (Vector2){GetScreenWidth() / 2 - 100, y+=60}, (Vector2){200, 50}))
    {
        gameDifficulty = GAME_DIFFICULTY_NORMAL;
        nextScene = GAME_SCENE_PLAY;
    }

    if (Button("Play hard", (Vector2){GetScreenWidth() / 2 - 100, y+=60}, (Vector2){200, 50}))
    {
        gameDifficulty = GAME_DIFFICULTY_HARD;
        nextScene = GAME_SCENE_PLAY;
    }

    if (Button("Settings", (Vector2){GetScreenWidth() / 2 - 100, y+=60}, (Vector2){200, 50}))
    {
        nextScene = GAME_SCENE_SETTINGS;
    }
}
#if defined(PLATFORM_WEB)
#include <emscripten.h>

// Function to store data in Local Storage
void StoreData(const char *key, const char *value) {
    EM_ASM_({
        localStorage.setItem(UTF8ToString($0), UTF8ToString($1));
    }, key, value);
}

// Function to retrieve data from Local Storage
const char* RetrieveData(const char *key) {
    return (const char*)EM_ASM_INT({
        var value = localStorage.getItem(UTF8ToString($0));
        if (value === null) {
            return 0;
        }
        var lengthBytes = lengthBytesUTF8(value) + 1;
        var stringOnWasmHeap = _malloc(lengthBytes);
        stringToUTF8(value, stringOnWasmHeap, lengthBytes);
        return stringOnWasmHeap;
    }, key);
}
#else
void StoreData(const char *key, const char *value) {}
const char* RetrieveData(const char *key) { return 0; }
#endif

uint32_t RetrieveUInt32(const char *key)
{
    const char *value = RetrieveData(key);
    if (value)
    {
        uint32_t result = atoi(value);
        free((void*)value);
        return result;
    }
    return 0;
}

void StoreUInt32(const char *key, uint32_t value)
{
    char buffer[16];
    sprintf(buffer, "%d", value);
    StoreData(key, buffer);
}

void RetrieveFixedChar(const char *key, char *buffer, int size)
{
    const char *value = RetrieveData(key);
    if (value)
    {
        strncpy(buffer, value, size);
        free((void*)value);
    }
    else
    {
        buffer[0] = '\0';
    }
}

void StoreFixedChar(const char *key, const char *value)
{
    StoreData(key, value);
}


void LoadStorage()
{
    // ignore version as this is first version. Upgrades need to be backwards compatible
    for (int i = 0; i < 4; i++)
    {
        for (int j = 0; j < 8; j++)
        {
            char key[64];
            sprintf(key, "storage.highscore_%d_%d", i, j);
            storage.scores[i].highscores[j].score = RetrieveUInt32(key);
            sprintf(key, "storage.name_%d_%d", i, j);
            RetrieveFixedChar(key, storage.scores[i].highscores[j].name, 16);
            sprintf(key, "storage.date_%d_%d", i, j);
            RetrieveFixedChar(key, storage.scores[i].highscores[j].date, 16);
        }
    }
    storage.tutorialStep = RetrieveUInt32("storage.tutorialStep");
    storage.tutorialStepCount = RetrieveUInt32("storage.tutorialStepCount");
    storage.tutorialCompleted = RetrieveUInt32("storage.tutorialCompleted");
    storage.startups = RetrieveUInt32("storage.startups");
    StoreUInt32("storage.startups", storage.startups + 1);
}

void SaveStorage()
{
    StoreUInt32("storage.version", 1);
    for (int i = 0; i < 4; i++)
    {
        for (int j = 0; j < 8; j++)
        {
            char key[64];
            sprintf(key, "storage.highscore_%d_%d", i, j);
            StoreUInt32(key, storage.scores[i].highscores[j].score);
            sprintf(key, "storage.name_%d_%d", i, j);
            StoreFixedChar(key, storage.scores[i].highscores[j].name);
            sprintf(key, "storage.date_%d_%d", i, j);
            StoreFixedChar(key, storage.scores[i].highscores[j].date);
        }
    }
    StoreUInt32("storage.tutorialStep", storage.tutorialStep);
    StoreUInt32("storage.tutorialStepCount", storage.tutorialStepCount);
    StoreUInt32("storage.tutorialCompleted", storage.tutorialCompleted);
}

void LogStorage()
{
    TraceLog(LOG_INFO, "[storage] Storage log");
    for (int i = 0; i < 4; i++)
    {
        for (int j = 0; j < 8; j++)
        {
            TraceLog(LOG_INFO, "[storage] Highscore %d %d: %d %s %s", i, j, storage.scores[i].highscores[j].score, 
                storage.scores[i].highscores[j].name, storage.scores[i].highscores[j].date);
        }
    }
    TraceLog(LOG_INFO, "[storage] Tutorial: %d %d %d", storage.tutorialStep, storage.tutorialStepCount, storage.tutorialCompleted);
    TraceLog(LOG_INFO, "[storage] Startups: %d", storage.startups);
}

#include <time.h>
void StoreHighScore(PlayfieldScores scores)
{
    DifficultyScores *difficultyScores = &storage.scores[gameDifficulty];

    for (int i = 0; i < 8; i++)
    {
        if (scores.score > difficultyScores->highscores[i].score)
        {
            for (int j = 7; j > i; j--)
            {
                difficultyScores->highscores[j] = difficultyScores->highscores[j - 1];
            }
            difficultyScores->highscores[i].score = scores.score;
            time_t now = time(0);
            struct tm *tm = localtime(&now);
            strftime(difficultyScores->highscores[i].date, 16, "%Y-%m-%d", tm);
            difficultyScores->highscores[i].name[0] = '\0';
            SaveStorage();
            break;
        }
    }
}

void TutorialBubble(int step, const char *text, int x, int y, int width, int height)
{
    if (playfield.tutorialStep != step) return;
    Rectangle rect = {x, y, width, height};
    DrawRectangleRec(rect, (Color){200, 200, 200, 255});
    DrawRectangleLinesEx(rect, 2, BLACK);
    float textWidth = MeasureText(text, 20);
    DrawText(text, x + width * 0.5f - textWidth * 0.5f, y + 10, 20, BLACK);
}

void TutorialSetClicksBlocked(int step, int blocked)
{
    if (playfield.tutorialStep != step) return;
    isClickActionBlocked = blocked;
}

void TutorialProceedOnClick(int step)
{
    if (playfield.tutorialStep != step) return;
    if (IsMouseButtonReleased(MOUSE_LEFT_BUTTON)) {
        playfield.nextTutorialStep = step + 1;
    }
}

void TutorialHighlightCircle(int step, int x, int y, int radius)
{
    if (playfield.tutorialStep != step) return;
    float pulse = sinf(gameTime.time * 10.0f) * 0.25f + 1.0f;
    DrawCircleLines(x, y, radius * pulse, COLOR_FROM_HEX(0xff0000));
    DrawCircleLines(x, y, radius * pulse + 0.5f, COLOR_FROM_HEX(0xff0000));
    DrawCircleLines(x, y, radius * pulse + 1.0f, COLOR_FROM_HEX(0xff0000));
}

void TutorialSpawnBubble(int step, int x, int y, int level, int type, int nextLevel, int nextType)
{
    if (playfield.tutorialStep != step || !playfield.tutorialStepStart) return;
    Vector2 pos = PlayfieldScreenToSpawnPosition(&playfield, camera, 
            (Vector2){x, y});
    playfield.nextBubbleLevel = level;
    playfield.nextBubbleType = type;
    PlayfieldTryAddBubble(&playfield, pos);
    playfield.nextBubbleLevel = nextLevel;
    playfield.nextBubbleType = nextType;
}

void UpdateTutorialSystem_PlayTutorial()
{
    int step = 0;
    
    TutorialSetClicksBlocked(step, 1);
    TutorialBubble(step, "Welcome to my fishtank...", 20, 20, 400, 75);
    TutorialProceedOnClick(step);
    step++;

    TutorialSetClicksBlocked(step, 1);
    TutorialBubble(step, "It is very boring here.\nSo I invented this game:", 20, 20, 400, 75);
    TutorialProceedOnClick(step);
    step++;

    TutorialSetClicksBlocked(step, 1);
    TutorialBubble(step, "Please tap on the marked area ...", 20, 20, 400, 75);
    TutorialProceedOnClick(step);
    TutorialHighlightCircle(step, GetScreenWidth() / 2, GetScreenHeight() - 50, 20);
    step++;

    TutorialSpawnBubble(step, GetScreenWidth() / 2, GetScreenHeight() - 50, 1, 0, 1, 0);
    TutorialBubble(step, "Very good!\nNow do it again!", 20, 20, 400, 75);
    TutorialProceedOnClick(step);
    TutorialHighlightCircle(step, GetScreenWidth() / 2, GetScreenHeight() - 50, 20);
    step++;
    
    TutorialSpawnBubble(step, GetScreenWidth() / 2, GetScreenHeight() - 50, 1, 0, 1, 0);
    TutorialBubble(step, 
        "These two bubbles have\n"
        "the same size and color.\n"
        "When they touch, they will merge.", 20, 20, 400, 75);
    TutorialProceedOnClick(step);
    step++;
    // finished
    TutorialSetClicksBlocked(step, 0);
}

void UpdateTutorialSystem()
{
    if (currentScene == GAME_SCENE_PLAY && gameDifficulty == GAME_DIFFICULTY_TUTORIAL)
    {
        UpdateTutorialSystem_PlayTutorial();
    }

    if (playfield.nextTutorialStep != playfield.tutorialStep)
    {
        playfield.tutorialStepStart = 1;
    }
    else
    {
        playfield.tutorialStepStart = 0;
    }
    playfield.tutorialStep = playfield.nextTutorialStep;
}

int main(void)
{
    // Initialization
    //--------------------------------------------------------------------------------------
    const int screenWidth = 800;
    const int screenHeight = 450;

    InitWindow(screenWidth, screenHeight, "GGJ25 - Bubble Pop");

    SetTargetFPS(60);               // Set our game to run at 60 frames-per-second
    //--------------------------------------------------------------------------------------

    TraceLog(LOG_INFO, "loading shaders");
    bubbleOutlineShader = LoadShader(0, "data/bubble_outline.fs");
    // Get shader locations
    outlineSizeLoc = GetShaderLocation(bubbleOutlineShader, "outlineSize");
    outlineColorLoc = GetShaderLocation(bubbleOutlineShader, "outlineColor");
    textureSizeLoc = GetShaderLocation(bubbleOutlineShader, "textureSize");

    LoadStorage();
    LogStorage();

    bubbleModel = LoadModelFromMesh(GenMeshSphere(1.0f, 4, 24));
    // Main game loop
    while (!WindowShouldClose())    // Detect window close button or ESC key
    {
        if (bubbleFieldTexture.texture.width != GetScreenWidth() || bubbleFieldTexture.texture.height != GetScreenHeight())
        {
            UnloadRenderTexture(bubbleFieldTexture);
            bubbleFieldTexture = LoadRenderTexture(GetScreenWidth(), GetScreenHeight());
        }

        float dt = GetFrameTime();
        // clamp dt to prevent large time steps, e.g. when browser tab is inactive
        if (dt > 0.2f) dt = 0.2f;
        gameTime.time += dt;
        gameTime.deltaTime = dt;

        BeginDrawing();
        switch (currentScene)
        {
            default: UpdateSceneMenu(); break;
            case GAME_SCENE_PLAY: UpdateScenePlay(); break;
            case GAME_SCENE_GAMEOVER: UpdateSceneGameOver(); break;
        }

        UpdateTutorialSystem();

        switch (nextScene)
        {
            case GAME_SCENE_NONE: break;
            default: currentScene = nextScene; break;
            case GAME_SCENE_GAMEOVER:
                StoreHighScore(CalculatePlayfieldScores(&playfield));
                currentScene = GAME_SCENE_GAMEOVER;
                break;
            case GAME_SCENE_PLAY: 
                playfield = (Playfield){
                    .fieldSize = {230, 300},
                    .waterLevel = 200.0f,
                    .minHeight = 90.0f,
                    .maxHeight = 280.0f,
                };
                currentScene = GAME_SCENE_PLAY; break;
        }

        if (nextScene != GAME_SCENE_NONE)
        {
            playfield.tutorialStep = 0;
            playfield.nextTutorialStep = 0;
        }

        nextScene = GAME_SCENE_NONE;
        EndDrawing();
    }

    // De-Initialization
    //--------------------------------------------------------------------------------------
    CloseWindow();        // Close window and OpenGL context
    //--------------------------------------------------------------------------------------

    return 0;
}